// mm_srv client for Avisynth v2.5.x Copyright 2009 N099
//

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "mm_srv.h"
#include "avisynth.h"

#include "mm_cl_avs.h"

//utils
inline int ALIGN_NEXT(int n, int a)
{
	return (n + (a - 1)) & ~(a - 1);
}

//ctor
MMCLAvs::MMCLAvs(const char* filename, int fmt, bool interlaced, bool tff,
				 IScriptEnvironment* env) : IClip(), _hmms(NULL)
{
	try{
		//open server
		_hmms = open_mm_srv(filename, fmt, FRAME_ALIGN);
		if(!_hmms){
			env->ThrowError("mm_cl_avs: open_mm_srv failed!");
		}

		//set video info
		ZeroMemory(&_vi, sizeof(_vi));
		SetVideoInfo(interlaced, tff);
	}catch(...){
		CleanUp();
		throw;
	}
}

/// private functions

//resource clean up
void MMCLAvs::CleanUp()
{
	if(_hmms){
		close_mm_srv(_hmms);
		_hmms = NULL;
	}
}

//set vieo info
void MMCLAvs::SetVideoInfo(bool interlaced, bool tff)
{
	//this won't occur.
	if(!_hmms) return;

	//set info
	_vi.width           = _hmms->w;
	_vi.height          = _hmms->h;
	_vi.fps_numerator   = _hmms->rate;
	_vi.fps_denominator = _hmms->scale;
	_vi.num_frames      = _hmms->frame;
	_vi.pixel_type      = MM_FMT2VideoinfoFormat(_hmms->format);

	//no audio
	_vi.audio_samples_per_second = _vi.sample_type = _vi.nchannels = 0;
	_vi.num_audio_samples = 0;

	//image-type properties
	if(interlaced){
		_vi.image_type = tff ? VideoInfo::IT_TFF : VideoInfo::IT_BFF;
	}else{
		_vi.image_type = 0;
	}
}

//fourcc string to MM_FMT_X
int MMCLAvs::Str2MM_FMT(const char* str, bool interlaced)
{
	int ret = -1;

	if(0 == lstrcmpiA(str, "YUY2")){
		ret = MM_FMT_YUY2;
	}else if(0 == lstrcmpiA(str, "YV12")){
		ret = interlaced ? MM_FMT_YV12I : MM_FMT_YV12;
	}else if(0 == lstrcmpiA(str, "RGB24")){
		ret = MM_FMT_RGB24;
	}

	return ret;
}

//MM_FMT to VideoInfo::CS_X
int MMCLAvs::MM_FMT2VideoinfoFormat(unsigned long fmt)
{
	int ret = VideoInfo::CS_UNKNOWN;

	switch(fmt){
		case MM_FMT_RGB24:
		case MM_FMT_RGB24R:
			ret = VideoInfo::CS_BGR24;
			break;
		case MM_FMT_YUY2:
			ret = VideoInfo::CS_YUY2;
			break;
		case MM_FMT_YV12:
		case MM_FMT_YV12I:
			ret = VideoInfo::CS_YV12;
			break;
		case MM_FMT_YC48: //muri
		default:
			ret = VideoInfo::CS_UNKNOWN;
	}

	return ret;
}

///public functions

//impl IClip
PVideoFrame __stdcall
MMCLAvs::GetFrame(int n, IScriptEnvironment* env)
{
	PVideoFrame pvf = env->NewVideoFrame(_vi);
	BYTE* buf = static_cast<BYTE*>(get_frame_mm_srv(_hmms, n));

	//get_frame failed / not opened
	if(!buf || !_hmms){
		//filled with gray
		FillMemory(pvf->GetWritePtr(), 128,
			pvf->GetPitch() * (pvf->GetHeight() + pvf->GetHeight(PLANAR_U)));
		return pvf;
	}

	//copy to VideoFrame
	// an image in "buf" is aligned by FRAME_ALIGN bytes.
	if(_vi.IsPlanar()){
		//YV12
		const int yh = pvf->GetHeight(PLANAR_Y);
		const int yw = pvf->GetRowSize(PLANAR_Y);
		const int ch = pvf->GetHeight(PLANAR_V);
		const int cw = pvf->GetRowSize(PLANAR_V);
		const int s  = _hmms->pitch;

		//Y
		env->BitBlt(pvf->GetWritePtr(PLANAR_Y), pvf->GetPitch(PLANAR_Y), buf, s, yw, yh);
		//V
		buf += s * yh;
		env->BitBlt(pvf->GetWritePtr(PLANAR_V), pvf->GetPitch(PLANAR_V), buf, s/2, cw, ch);
		//U
		buf += (s/2) * ch;
		env->BitBlt(pvf->GetWritePtr(PLANAR_U), pvf->GetPitch(PLANAR_U), buf, s/2, cw, ch);
	}else{
		//RGB,YUY2
		env->BitBlt(pvf->GetWritePtr(), pvf->GetPitch(),
			buf, _hmms->pitch, pvf->GetRowSize(), pvf->GetHeight());
	}

	return pvf;
}

//static class factory
AVSValue __cdecl
MMCLAvs::CreateMMCLAvs(AVSValue args, void*, IScriptEnvironment* env)
{
	if(args[0].ArraySize() != 1){
		env->ThrowError("mm_cl_avs: Only 1 filename currently supported!");
	}

	//get arguments
	const char* filename  = args[0][0].AsString();
	const bool interlaced = args[2].AsBool(false);
	const bool tff        = args[3].AsBool(true);

	//pixel type
	int fmt = MM_FMT_YUY2;
	if(args[1].Defined()){
		int tmp = Str2MM_FMT(args[1].AsString(), interlaced);
		if(tmp < 0){
			env->ThrowError("mm_cl_avs: pixel_type must be \"YUY2\", \"YV12\" or \"RGB24\"!");
		}
		fmt = (tmp > 0) ? tmp : MM_FMT_YUY2;
	}

	//create instance
	return new MMCLAvs(filename, fmt, interlaced, tff, env);
}

//export entry
extern "C" __declspec(dllexport) const char*
__stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	env->AddFunction("mm_cl_avs", "s+[pixel_type]s[interlaced]b[tff]b",
		MMCLAvs::CreateMMCLAvs, 0);
	return "mm_cl_avs";
}
